#
# Copyright (C) 2012 - present Instructure, Inc.
#
# This file is part of Canvas.
#
# Canvas is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, version 3 of the License.
#
# Canvas is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

define [
  'require'
  'compiled/models/FilesystemObject'
  'vendor/backbone-identity-map'
  'compiled/collections/PaginatedCollection'
  'compiled/collections/FilesCollection'
  'compiled/util/natcompare'
  'jsx/shared/helpers/urlHelper'
], (require, FilesystemObject, identityMapMixin, PaginatedCollection, FilesCollection, natcompare, urlHelper) ->


  Folder = identityMapMixin class __Folder extends FilesystemObject

    defaults:
      'name' : ''

    initialize: (options) ->
      @contentTypes ||= options?.contentTypes
      @useVerifiers ||= options?.useVerifiers
      @setUpFilesAndFoldersIfNeeded()
      @on 'change:sort change:order', @setQueryStringParams
      super

    url: ->
      if @isNew()
        super
      else
        "/api/v1/folders/#{@id}"

    parse: (response) ->
      json = super
      @contentTypes ||= response.contentTypes
      @useVerifiers ||= response.useVerifiers
      @setUpFilesAndFoldersIfNeeded()

      @folders.url = response.folders_url
      @files.url   = response.files_url

      json

    setUpFilesAndFoldersIfNeeded: ->
      unless @folders
        @folders = new FoldersCollection [], parentFolder: this
      unless @files
        @files = new FilesCollection [], parentFolder: this

    getSubtrees: ->
      @folders

    getItems: ->
      @files

    expand: (force=false, options={}) ->
      @isExpanded = true
      @trigger 'expanded'
      return $.when() if @expandDfd || force
      @isExpanding = true
      @trigger 'beginexpanding'
      @expandDfd = $.Deferred().done =>
        @isExpanding = false
        @trigger 'endexpanding'

      selfHasntBeenFetched = @folders.url is @folders.constructor::url or @files.url is @files.constructor::url
      fetchDfd = @fetch() if selfHasntBeenFetched || force
      $.when(fetchDfd).done =>
        foldersDfd = @folders.fetch() unless @get('folders_count') is 0
        filesDfd = @files.fetch() if (@get('files_count') isnt 0) and !options.onlyShowSubtrees
        $.when(foldersDfd, filesDfd).done(@expandDfd.resolve)

    collapse: ->
      @isExpanded = false
      @trigger 'collapsed'

    toggle: (options) ->
      if @isExpanded
        @collapse()
      else
        @expand(false, options)

    previewUrl: ->
      if @get('context_type') in ['Course', 'Group']
        "/#{ @get('context_type').toLowerCase() + 's' }/#{ @get('context_id') }/files/{{id}}/preview"

    isEmpty: ->
      !!(@files.loadedAll and (@files.length is 0)) and (@folders.loadedAll and (@folders.length is 0))

    # `full_name` will be something like "course files/some folder/another".
    # For routing in the react app in the browser, we want something that will take that "course files"
    # out. because urls will end up being /courses/2/files/folder/some folder/another
    EVERYTHING_BEFORE_THE_FIRST_SLASH = /^[^\/]+\/?/
    filesEnv = null
    urlPath: ->
      relativePath = (@get('full_name') or '').replace(EVERYTHING_BEFORE_THE_FIRST_SLASH, '')
      relativePath = urlHelper.encodeSpecialChars(relativePath)
      relativePath = relativePath.split('/').map((component) ->
        encodeURIComponent(component)
      ).join('/')

      filesEnv ||= require('compiled/react_files/modules/filesEnv') # circular dep

      # when we are viewing all files we need to pad the context_asset_string on the front of the url
      # so it would be something like /files/folder/users_1/some/sub/folder
      if filesEnv.showingAllContexts
        assetString = "#{@get('context_type')?.toLowerCase()}s_#{@get('context_id')}"
        relativePath = assetString + '/' + relativePath

      relativePath

    @resolvePath = (contextType, contextId, folderPath) ->
      folderPath = urlHelper.decodeSpecialChars(folderPath)

      url = "/api/v1/#{contextType}/#{contextId}/folders/by_path#{folderPath}"
      $.getJSON(url).pipe (folders) ->
        folders.map (folderAttrs) ->
          new Folder(folderAttrs, {parse: true})

    getSortProp = (model, sortProp) ->
      # if we are sorting by name use 'display_name' for files and 'name' for folders.
      if sortProp is 'name' and not (model instanceof Folder)
        model.get('display_name')
      else if sortProp is 'user'
        model.get('user')?.display_name || ''
      else if sortProp is 'usage_rights'
        model.get('usage_rights')?.license_name || ''
      else
        model.get(sortProp)

    ##
    # Special sorter for handling sorting with special properties
    # It's been enhanced to sort naturally when certain sortProps
    # are used.
    childrenSorter: (sortProp='name', sortOrder='asc', a, b) ->
      # Only use natural mode for instances we expect strings in.
      naturalMode = sortProp in ['name', 'user', 'usage_rights']

      # Get actual values for the properties we are sorting by.
      a = getSortProp(a, sortProp)
      b = getSortProp(b, sortProp)

      if naturalMode
        res = natcompare.strings(a, b)
      else
        res = if a is b
                0
              else if a > b or a is undefined
                1
              else if a < b or b is undefined
                -1
              else
                throw new Error("wat? error sorting")

      res = 0 - res if sortOrder is 'desc'
      res

    children: ({sort, order}) ->
      (@folders.toArray().concat @files.toArray()).sort(@childrenSorter.bind(null, sort, order))








  # FoldersCollection is defined inside of this file, and not where it
  # should be, because RequireJS sucks at figuring out circular dependencies.
  # 'compiled/collections/FoldersCollection' just grabs this and re-exports it.
  Folder.FoldersCollection = class FoldersCollection extends PaginatedCollection
    @optionProperty 'parentFolder'

    model: Folder

    parse: (response) ->
      if response
        response.forEach (folder) =>
          folder.contentTypes = @parentFolder.contentTypes
          folder.useVerifiers = @parentFolder.useVerifiers
      super





  return Folder
