module Types
  CourseType = GraphQL::ObjectType.define do
    name "Course"

    implements GraphQL::Relay::Node.interface

    global_id_field :id
    field :_id, !types.ID, "legacy canvas id", property: :id
    field :name, !types.String
    field :courseCode, types.String,
      "course short name",
      property: :course_code
    field :state, !CourseWorkflowState,
      property: :workflow_state

    connection :assignmentsConnection do
      type AssignmentType.connection_type
      resolve -> (course, _, ctx) {
        Assignments::ScopedToUser.new(course, ctx[:current_user]).scope
      }
    end

    connection :sectionsConnection do
      type SectionType.connection_type
      resolve -> (course, _, ctx) {
        course.active_course_sections.
          order(CourseSection.best_unicode_collation_key('name'))
      }
    end

    connection :usersConnection do
      type UserType.connection_type
      resolve ->(course, _, ctx) {
        if course.grants_any_right?(ctx[:current_user], ctx[:session],
            :read_roster, :view_all_grades, :manage_grades)
          UserSearch.scope_for(course, ctx[:current_user], {})
        else
          nil
        end
      }
    end

    connection :gradingPeriodsConnection, GradingPeriodType.connection_type do
      resolve ->(course, _, _) {
        GradingPeriod.for(course).order(:start_date)
      }
    end

    connection :submissionsConnection, SubmissionType.connection_type do
      description "all the submissions for assignments in this course"

      argument :studentIds, !types[!types.ID], "Only return submissions for the given students.",
        prepare: GraphQLHelpers.relay_or_legacy_ids_prepare_func("User")
      argument :orderBy, types[SubmissionOrderInputType]

      resolve ->(course, args, ctx) {
        current_user = ctx[:current_user]
        session = ctx[:session]
        user_ids = args[:studentIds].map(&:to_i)

        if course.grants_any_right?(current_user, session, :manage_grades, :view_all_grades)
          # TODO: make a preloader for this???
          allowed_user_ids = course.apply_enrollment_visibility(course.all_student_enrollments, current_user).pluck(:user_id)
          allowed_user_ids &= user_ids
        elsif course.grants_right?(current_user, session, :read_grades)
          allowed_user_ids = user_ids & [current_user.id]
        else
          allowed_user_ids = []
        end

        submissions = Submission.active.joins(:assignment).where(
          user_id: allowed_user_ids,
          assignment_id: course.assignments.published
        ).where.not(workflow_state: "unsubmitted")

        (args[:orderBy] || []).each { |order|
          submissions = submissions.order(order[:field] => order[:direction])
        }

        submissions
      }
    end

    field :permissions, CoursePermissionsType do
      description "returns permission information for the current user in this course"
      resolve ->(course, _, ctx) {
        Loaders::CoursePermissionsLoader.for(
          course,
          current_user: ctx[:current_user], session: ctx[:session]
        )
      }
    end
  end

  SubmissionOrderInputType = GraphQL::InputObjectType.define do
    name "SubmissionOrderCriteria"
    argument :field, !GraphQL::EnumType.define {
      name "SubmissionOrderField"
      value "_id", value: "id"
      value "gradedAt", value: "graded_at"
    }
    argument :direction, GraphQL::EnumType.define {
      name "OrderDirection"
      value "ascending", value: "ASC"
      value "descending", value: "DESC"
    }
  end

  CourseWorkflowState = GraphQL::EnumType.define do
    name "CourseWorkflowState"
    description "States that Courses can be in"
    value "created"
    value "claimed"
    value "available"
    value "completed"
    value "deleted"
  end
end
